#include "Application.h"
#include "Pad.h"
#include <iostream>
#include "Config.h"
#include "Socket.h"
#include "Command.h"
#include "utils.h"
#include "PadManager.h"
#include "MidiPort.h"
#include <sstream>
#include <fstream>



void CommandLog::write(ISocket& sock, const std::string& append)
{

	sock.write(::str(error)+" "+message+"\n");
	std::string out=" ", _str=str()+append;
	int len = _str.size();
	for(int i=0; i<len; i++){
		out+=_str[i];
		if(_str[i]=='\n') out+=" ";
	}
	sock.write(out);
}

void add_commands(Application& app);
Application::Application()
{
	m_pad_last_id=-1;
	m_server=NULL;
	int port = Config::get_int("server.port");
	const std::string& type = Config::get_string("server.type");
	const std::string& addr = Config::get_string("server.address");

	if(type=="af_unix")
	{
		UnixServer* p = new UnixServer();
		m_server=p;
		p->listen(addr);
	}
	else if(type=="inet")
	{
		InetServer* p = new InetServer();
		m_server=p;
		p->listen(port);
	}
	else
	{
		StdioServer* p = new StdioServer();
		m_server=p;
	}
	add_commands(*(Application*)this);
	_auto_load_pad();
}

Application::~Application()
{
	delete m_server;
	int i, s=m_callbacks.size();
	for(i=0; i<s; i++)
		delete m_callbacks[i];
	m_callbacks.clear();
}

void Application::process_next_command(int n)
{
	if(m_server->get_type()==Socket::STDIO) std::cerr << "> ";
	Command::CommandReturn do_continue = Command::CONTINUE;
	ISocket* sock = NULL;
	while(do_continue!=Command::EXIT){
		if(!sock || !sock->is_open() || do_continue==Command::CLOSE || !Config::get_bool("server.keep-alive")){
			if(sock) delete sock;
			sock=m_server->accept();
		}

		if(m_server->get_type()!=Socket::STDIO)
			sock->write("\n> ");
		Command c;
		try{
			c=Command::parse(sock);
		} catch (ParseError p) {
			SEND(*sock, p.error);
		}

		if(c.get_name()!="")do_continue=_call(c, *sock);

		// fermeture de la socket si keep alive n'est pas activé et si ce n'est pas stdio
		if(m_server->get_type()!=Socket::STDIO
		   && !Config::get_bool("server.keep-alive")){
			sock->close();
		}

		// Affichage du prompt ("> ") si la socket est stdio (pour les autres, c'est fait en début de boucle)
		if(!do_continue && m_server->get_type()==Socket::STDIO) sock->write("> ");
		if(n<0) n--;

	}
}

void Application::process()
{
	process_next_command(-1);
}

Pad& Application::operator[](int id)
{
	if(id>=0 && id<m_pads.size())
		return *m_pads[id].second;
	return *(Pad*)NULL;
}

Pad& Application::operator[](const std::string& name)
{
	return get_pad(name);
}

Pad& Application::set_pad(int x)
{
	int i, s=m_pads.size();
	for(i=0; i<s; i++)
	{
		if(m_pads[i].first==x)
			break;
	}
	if(i==s){
		THROW(NotFoundError, "Le pad id(" << x << ") est introuvable");
	}else{
		m_current_index=x;
		m_current=&get_pad(x);
	}
	return *m_current;
}

Pad& Application::set_pad(const std::string& name)
{
	int i, s=m_pads.size();
	for(i=0; i<s; i++)
	{
		if(m_pads[i].second->get_name()==name)
			break;
	}
	if(i==s){
		std::cerr << "Le pad '" << name << "' est introuvable\n";
	}else{
		m_current=&get_pad(name);
		m_current_index=i;
	}
	return *m_current;
}

Pad& Application::get_pad(const std::string& name)
{
	if(name=="") return *m_current;
	int i, s=m_pads.size();
	for(i=0; i<s; i++)
		if(m_pads[i].second->get_name()==name)
			return *m_pads[i].second;
	std::cerr << "Pad '"<< name << " ' introuvable\n";
	return *(Pad*)NULL;
}

Pad& Application::get_pad(int id)
{
	if(id<0) return *m_current;
	int i, s=m_pads.size();
	for(i=0; i<s; i++)
		if(m_pads[i].first==id)
			return *m_pads[i].second;
	std::cerr << "Pad id("<< id << " ) introuvable\n";
	THROW(NotFoundError, "Pad not found");
}

int Application::add_pad(Pad* p)
{
	std::pair<int, Pad*> pp(++m_pad_last_id, p);
	m_pads.push_back(pp);
	return m_pads[m_pads.size()-1].first;
}

int Application::add_pad(const std::string& name)
{
	Pad* pad = new Pad(PadManager::get_pad(name));
	std::pair<int, Pad*> pp(++m_pad_last_id, pad);
	m_pads.push_back(pp);
	return m_pads[m_pads.size()-1].first;
}

bool Application::has_pad(const std::string& name)
{
	int i, s=m_pads.size();
	for(i=0; i<s; i++)
		if(m_pads[i].second->get_name()==name)
			return true;
	return false;
}

void Application::remove_pad(int id)
{
	int i, s=m_pads.size()-1;
	for(i=s; i>=0; i--)
		if(m_pads[i].first==id){
			m_pads.erase(m_pads.begin()+i);
			return;
		}
}

void Application::remove_pad(const std::string& name)
{
	int i, s=m_pads.size()-1;
	for(i=s; i>=0; i--)
		if(m_pads[i].second->get_name()==name)
			m_pads.erase(m_pads.begin()+i);
}

void Application::add_command(const std::string& name, CommandCallback c, const std::string& h)
{
	m_callbacks.push_back(new CommandHandler(name, c, h));
}

Command::CommandReturn Application::_call(Command& c, ISocket& sock)
{
	int i, s=m_callbacks.size();
	for(i=0; i<s; i++)
	{
		if(c.get_name()==m_callbacks[i]->name)
		{
			CommandLog log;
			try {
				Command::CommandReturn ret = m_callbacks[i]->command(*(Application*)this, *m_current, c.get_args(), log);
				log.write(sock);
				return ret;
			} catch(Error err){
				log.set(err.code, err.type);
				log.write(sock, err.error);
			} catch(SocketError err){
				log.set(-1024, err.type);
				log.write(sock, err.error);
			}
			return Command::CONTINUE;
		}
	}
	CommandLog log;
	log.set(1, "Commande introuvable");
	log << "Command '"<< c.get_name() << "' introuvable";
	log.write(sock);
	return Command::CONTINUE;
}

int Application::get_id(const std::string& name){
	int i, s=m_pads.size();
	for(i=0; i<s; i++)
		if(name==m_pads[i].second->get_name())
			return m_pads[i].first;
	return -1;
}

void Application::help(CommandLog& out)
{
	int i, s=m_callbacks.size();
	for(i=0; i<s; i++)
	{
		out << m_callbacks[i]->name+" : " << m_callbacks[i]->help+"\n";
	}
}


bool Application::load_pad_configuration(const std::string& p)
{
	if(Fs::is_file(p)){
		std::ifstream ifs(p);
		std::string padname;
		Json::Value c;
		ifs >> c;
		if(c.isObject() && c.isMember("pad") && c["pad"].isString())
			padname=c["pad"].asString();

		Pad& pad = get_pad(padname);
		if(pad){
			pad.load_configuration(p);
		}else{
			int x = add_pad(padname);
			Pad* pad = &get_pad(x);
			if(pad){
				pad->load_configuration(p);
			}else{
				THROW(UnknownError, "Impossible d'ajouter le pad '"<<padname<<"'");
			}
		}
	}
	return true;
}

bool Application::load_pad_definition(const std::string& padname)
{
	Pad& pad = get_pad(padname);
	if(!pad){
		int x = add_pad(padname);
		Pad* pad = &get_pad(x);
		if(!pad){
			THROW(UnknownError, "Impossible d'ajouter le pad '"<<padname<<"'");
		}
	}
	return true;
}


void Application::_auto_load_pad()
{

	std::vector<std::string> ports = MidiPortIn::get_ports();
	int n = PadManager::count(), nports= ports.size();

	for(int i=0; i<n; i++){
		PadEntry* p = PadManager::get_info_at(i);
		if(!has_pad(p->name)){
			for(int j=0; j<nports; j++){
				if(start_with(ports[j], p->in_port)){
					int x = add_pad(p->name);
					Pad& p = get_pad(x);
					p.connect(true, true);

				}
			}
		}

	}
}

