#include "PadDefinition.h"
#include <fstream>

#include "InputDefinition.h"

#include "Button.h"
#include "Controller.h"

AbsInput* __pad_default_button(PadDefinition* p, void * data)
{
    return new Button(p);
}


AbsInput* __pad_default_controller(PadDefinition* p, void * data)
{
    return new Controller(p);
}

Button* PadDefinition::instanciate_new_button() const
{
	return dynamic_cast<Button*>(m_create_button((PadDefinition*)this, m_create_button_data));
}

Controller* PadDefinition::instanciate_new_controller() const
{
	return dynamic_cast<Controller*>(m_create_controller((PadDefinition*)this, m_create_controller_data));
}


AbsInput* AbsInput::from_json(PadDefinition* p, Json::Value& v)
{
    if(v.isObject())
    {
        std::string s = v["type"].asString();
        InputDefinition* id = p->get_input_def(s);
        if(!id)
			 THROW(NotFoundError,"JsonErreur: InputDefinition '"+s+"' est inconnue");
        return id->instanciate(p, v);
    }
    else if(v.isNull()) return NULL;
	THROW(ParseError, "Erreur les données JSon n'est pas un objet");
}

std::vector<AbsInput*>& PadDefinition::get_matrix()
{
    return m_matrix;
}


PadDefinition::PadDefinition()
{
    m_iport=new MidiPortIn(m_name+" Controller");
    m_cport=new MidiPortOut(m_name+" Controller");
    m_oport=new MidiPortOut(m_name+" Controller");
    m_iport->set_listener(this);
    m_iport->openVirtualPort("Device in");
    m_cport->openVirtualPort("Device out");
    m_oport->openVirtualPort("Midi Out");
    m_oport_open=true;
    m_height=m_width;
    m_error_listener=NULL;
    m_port_listener=NULL;
    m_create_button=__pad_default_button;
    m_create_controller=__pad_default_controller;
    m_create_button_data=m_create_controller_data=NULL;
}
/**
    ACTIONS_DEFINITION -> [message : string, args, ...] | [[message : string, args, ...], ...]

    INPUT_DEFINTIONS -> {
            "type": string # "BUTTON" or "CONTROLLER"
            "actions" : {
                {action_id}: ACTIONS_DEFINITION
            }
            "shape" : string # Name of shape //to implement
        }

    INPUT -> {
            "type": string # input_defintion_id,
            "channel" : int # fixed channel or -1 (to replace)
            "key" : int # fixed key or -1 (to replace)
            "velocity" : int # fixed velocity or -1 (to replace)
            "locked" : string or null # @action_id
        }



    PAD -> {
        "name" : string #name of pad
        "width": int #width of matrix input,
        "height": int #height of matrix input,
        "ports": {
            "input": string #Name of input port
            "output": string #Name of output port
        },
        "inputs" : {
            {input_defintion_id_N} : INPUT_DEFINTIONS
        }
        "matrix" : [INPUT, INPUT, INPUT, ..... ] # x*y
    }
*/
#include "PadManager.h"

PadDefinition::PadDefinition(Json::Value v)
{
    read_conf(v);
}
PadDefinition::PadDefinition(const std::string& v)
{
    if(!PadManager::has_pad(v))
    {
		THROW(NotFoundError, "Pad '"+v+"' est introuvable");
    }
    Json::Value pad = PadManager::read_pad(v);
    read_conf(pad);
}

void PadDefinition::read_conf(const std::string& file)
{
    std::ifstream  fd(std::string(PAD_DEFINITION_DIR)+"/"+file);
	Json::Value val;
    fd >> val;
    read_conf(val);
}

void PadDefinition::clear()
{
    send_all("clear");
}

void PadDefinition::send_all(const std::string& name)
{
    int l=m_width*m_height;
    for(int i=0; i<l; i++){
        AbsInput* input = m_matrix[i];
        if(input){
            input->control(name, true);
        }
    }
}

void PadDefinition::read_conf(Json::Value& v)
{
	if(!v.isObject()) THROW(MalformedError, "JSonErreur: Pad doit être un objet");

    m_error_listener=NULL;
    m_port_listener=NULL;
    m_name = v["name"].asString();
    m_width = v["width"].asInt();
    m_height = v["height"].asInt();
    m_port_in_name = v["ports"]["input"].asString();
    m_port_out_name = v["ports"]["output"].asString();
    m_iport=new MidiPortIn(m_name+" Controller");
    m_cport=new MidiPortOut(m_name+" Controller");
    m_oport=new MidiPortOut(m_name+" Controller");
    m_iport->set_listener(this);
    m_iport->openVirtualPort("Device in");
    m_cport->openVirtualPort("Device out");
    m_oport->openVirtualPort("Midi Out");


    Json::Value root = v["inputs"];
    for( Json::Value::const_iterator itr = root.begin() ; itr != root.end() ; itr++ ) {
        std::string k = itr.key().asString();
		Json::Value v = *itr;
        m_input_def.push_back(new InputDefinition(k,v));

    }
    m_create_button=__pad_default_button;
    m_create_controller=__pad_default_controller;
    m_create_button_data=m_create_controller_data=NULL;
    root = v["matrix"];
    int s = root.size();
    for(int i=0; i<s; i++)
    {
        AbsInput* b = AbsInput::from_json(this, root[i]);
        if(b)
        {
            b->x = i%m_width;
            b->y = i/m_width;
            b->index = i;
        }
        m_matrix.push_back(b);
    }
}

PadDefinition::~PadDefinition()
{
    delete m_iport;
    delete m_cport;
    delete m_oport;
}


InputDefinition* PadDefinition::get_input_def(const std::string& s)
{
    int size = m_input_def.size();
    for(int i=0; i<size; i++)
    {
        if(m_input_def[i]->m_name==s) return m_input_def[i];
    }
    return NULL;
}

void PadDefinition::connect_to_device(bool input, bool output)
{
    int nport;
    std::string portName;

    if(input && !m_iport_open){
        nport=m_iport->getPortCount();
        for(int i=0; i<nport; i++)
        {
            portName = m_iport->getPortName(i);
            if(start_with(portName, m_port_in_name))
            {
				//std::cout << "Find input port : "<<portName << "\n";
                m_iport->openPort(i, "Device in");
                m_iport_open=true;
                break;
            }
        }
    }
    if(output && !m_cport_open){
        nport = m_cport->getPortCount();
        for(int i=0; i<nport; i++)
        {
            portName = m_cport->getPortName(i);
            if(start_with(portName, m_port_out_name))
            {
                std::cout << "Find output port : "<<portName << "\n";
                m_cport->openPort(i, "Device out");
                m_cport_open=true;
                break;
            }
        }
    }
}

void PadDefinition::set_error_listener(IMidiPortErrorListener * p)
{
    m_error_listener=p;
}

void PadDefinition::set_port_listener(IMidiPortListener * p )
{
    m_port_listener=p;
}

void PadDefinition::on_error(RtMidiError::Type type, const std::string &errorText)
{
    if(m_error_listener){
        m_error_listener->on_error(type, errorText);
    }
    else
    {
        std::cerr << "Erreur : " << errorText << "\n";
		THROW(MidiError, "Erreur Midi: "+errorText);
    }
}

void PadDefinition::on_new_message(double ts, MidiMessage* m)
{
    if(m_port_listener){
        m_port_listener->on_new_message(ts, m);
    }
    else
    {
        _dispatch_message(ts, m);
    }
}

void PadDefinition::_dispatch_message(double ts, MidiMessage* m)
{
    AbsInput* in = get_by_message(m);
    if(in) in->on_input(ts, m);

}

AbsInput* PadDefinition::get_by_message(const MidiMessage* msg)
{
    InputType type = BUTTON;
    switch(msg->type)
    {
        case MidiMessage::CONTROLLER_CHANGE:
            type = CONTROLLER;
        case MidiMessage::NOTE_ON:
        case MidiMessage::NOTE_OFF:
        {
            const Note *n = dynamic_cast<const Note*>(msg);
            return get_by_channel_key(type, n->channel, n->key);
        }

    }
    return NULL;
}

AbsInput* PadDefinition::input_at(int x, int y)
{
    if(y<0) return m_matrix[x];
    return m_matrix[y*m_width+x];
}

AbsInput* PadDefinition::get_by_channel_key(InputType type, int channel, int key)
{
    int s = m_matrix.size();
    for(int i=0; i<s; i++)
    {
        AbsInput* b = m_matrix[i];

        if(b && b->type==type && b->realchannel==channel && b->realkey==key)
            return b;
    }
    return NULL;
}

PadDefinition* PadDefinition::from_file(const std::string& str)
{
    std::ifstream  fd(str);
    Json::Value val;
    printf("IsOpen : %d\n", fd.is_open());
    fd >> val;
	return new PadDefinition(val);
}

PadDefinition* PadDefinition::from_name(const std::string& str)
{
    std::ifstream  fd(std::string(PAD_DEFINITION_DIR)+"/"+str);
    Json::Value val;
    printf("IsOpen : %d\n", fd.is_open());
    fd >> val;
	return new PadDefinition(val);
}
